<a class='fenicsnav' href='{{URL}}/index.html'>Home</a>
&nbsp;
&nbsp;
<a class='fenicsnav' href='{{URL}}/install.html'>Installation</a>
&nbsp;
&nbsp;
<a class='fenicsnav' href='{{URL}}/python/demo'>Demos</a>
&nbsp;
&nbsp;
<a class='fenicsnav' href='{{URL}}/cpp'>C++ docs</a>
&nbsp;
&nbsp;
<a class='fenicsnav' href='{{URL}}/python'>Python docs</a>
